/////////////////////////////////////////////////////
// Nikon Camera Intervalometer
// (C) 2010 Jan Wagner <jan.wagner@iki.fi>
//
// Tested on Nikon D90 but should work on all Nikon 
// cameras supported by the Nikon remote ML-L3.
//
// Configuration is done via a single button.
//
// - first signal opens shutter indefinitely
// - second signal closes shutter
//
/////////////////////////////////////////////////////


/////////////////////////////////////////////////////
// For asm output:
// -S -fverbose-asm
//
// Fuse bits:
//   ATTINY13 configured for 4.0 MHz internal RC
//   CKSEL[3:0] = 0010 (4.8MHz)
//   SUT[1:0] = 10 (14CK+65 ms) slowly rising power
//   BODLEVEL[2:0] = 111  disable BOD
//   http://www.engbedded.com/fusecalc/
//   => -U lfuse:w:0x39:m -U hfuse:w:0xff:m
/////////////////////////////////////////////////////

// NOTE: TIMING FUDGE FACTORS USED!!
// See all calls to _delay_us() and the sleep_63ms()

/////////////////////////////////////////////////////
// The IR signal has a 38.4kHz modulation 
// that is a period of 26us with 13us on and 13us off
// 4.8 MHz means 13us = 13e-6*4.8e6 = 62.4 cycles
//    62 clock cycles for ModOn, 62 cycles for ModOff
// Thus, 31 cycles per one 26us modulation period
//
//    State   Duration   In 26us periods
// 1) ON  for 2000us     77
// 2) OFF for 27830us    1070
// 3) ON  for 390us      15
// 4) OFF for 1580us     61
// 5) ON  for 410us      16
// 6) OFF for 3580us     138
// 7) ON  for 400us      15
// 8) repeat once after 63200us
/////////////////////////////////////////////////////

/////////////////////////////////////////////////////
// Pin assignment ATTINY2313
// 4=gnd, 8=3.0V Vcc via switch
// 1=nRESET
// 2=PB3=IR LED out
// 3=PB4=shutter button to GND
// 7=SCL
// 6=MISO
// 5=MOSI
/////////////////////////////////////////////////////

/////////////////////////////////////////////////////
// USBTinyISP cable:			attiny
//   valk RST					1
//   kelt SCK					19
//   or   MISO					18
//   pun  MOSI					17
//   mus  GND					10
//   sin  +5V Vcc				20
/////////////////////////////////////////////////////

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/io.h>
#include <util/delay.h>

#define IR_LED_PIN  PB3
#define SHUTTER_PIN PB4
#define DBG_LEARN_PIN PB2
#define DBG_SEND_PIN  PB1
#define DBG_HALFSEC_PIN PB0

#define PORTB_DDR_MASK (_BV(IR_LED_PIN) | _BV(DBG_LEARN_PIN) | _BV(DBG_SEND_PIN) | _BV(DBG_HALFSEC_PIN))
#define DEBUG 0

unsigned char bulb_mode = 1;

unsigned int on_time = 0;
unsigned int pause_time = 0;

// System clock 4.8e6 Hz, Timer0 divider 1024, count from y to 255
// 1e3/(4.8e6/(1024*(255-20))) = 50.1333ms
// and output ticks 200ms means x4
#define TCNT0_BASE       20
#define TICKS_OUT_RATIO  4
volatile unsigned char ticks_200ms = 0;
volatile unsigned char ticks_tmp = 0; // 0..(TICKS_OUT_RATIO-1)

////////////////////////////////////////////////////////////
// PLATFORM SETUP
////////////////////////////////////////////////////////////

void init()
{
   MCUCR = 0; // PUD=0 so pull-up is enabled
   wdt_disable();
   set_sleep_mode(SLEEP_MODE_IDLE); // can wake from Timer Overflow
	ACSR = _BV(ACD); // disable analog comparator
	PORTB = 0;
	DDRB = 0;
	DDRB = PORTB_DDR_MASK;
	PORTB |= _BV(SHUTTER_PIN); // pull-up

	/* Configure system tick timer */
	cli();
	ticks_200ms = 0;
	ticks_tmp = 0;
   TCNT0 = TCNT0_BASE;
	TCCR0A = 0; // compare op disabled, WGM{00|0}=normal count up, overflow at 0xFF
   TCCR0B = _BV(CS02)|_BV(CS00); // rate '101'=clk_io/1024=4688 Hz
   // 4.8e6 Hz / (1024 * 255) = 18.3 Hz = 54.4 ms
   TIFR0  |= _BV(TOV0);  // '1' to clear TOV0 flag
   TIMSK0 |= _BV(TOIE0); // overflow ie
   sei();
}

void sleep_200ms()
{
	unsigned char tstart = ticks_200ms;
	while (tstart == ticks_200ms) {
	   // switch all pins to input, sleep, restore IR as output
	   DDRB = 0;
	   sleep_enable();
		DDRB = PORTB_DDR_MASK;
	}
}

////////////////////////////////////////////////////////////
// SYSTEM TICKS AT 200ms
////////////////////////////////////////////////////////////

ISR(TIM0_OVF_vect)
{
   TCNT0 = TCNT0_BASE;
	ticks_tmp++;
	if (ticks_tmp >= TICKS_OUT_RATIO) {
		ticks_tmp = 0;
      ticks_200ms++;
		if (PORTB & _BV(DBG_HALFSEC_PIN)) {
			PORTB &= ~(_BV(DBG_HALFSEC_PIN));
      } else {
			PORTB |= _BV(DBG_HALFSEC_PIN);
		}
	}
}

////////////////////////////////////////////////////////////
// MODULATED '0' SIGNAL
////////////////////////////////////////////////////////////

static inline void one_period_OFF()
{
   // wait 104 cycles @ 4 MHz (104/4e6 = 26us)
	PORTB &= ~(_BV(IR_LED_PIN));
	_delay_us(26+2*1);
   return;
}


////////////////////////////////////////////////////////////
// MODULATED '1' SIGNAL
////////////////////////////////////////////////////////////

static inline void one_period_ON()
{
   // wait 104 cycles @ 4 MHz (104/4e6 = 26us)
	// 1cyc=1/4e6=0.25us
	PORTB |= _BV(IR_LED_PIN);
   _delay_us(13+1); // 3cyc/iter, 13*3*0.25us=9.75us
	PORTB &= ~(_BV(IR_LED_PIN));
   _delay_us(13+1);
   return;
}


////////////////////////////////////////////////////////////
// SHUTTER RELEASE CODE SEQUENCE (must be repeated twice)
////////////////////////////////////////////////////////////

void shutter_release_1seq()
{
	unsigned char i;
	#if DEBUG
   PORTB |= _BV(DBG_SEND_PIN);
	#endif

   //    State   Duration   In 26us periods
   // 1) ON  for 2000us     77
   // 2) OFF for 27830us    1070
   // 3) ON  for 390us      15
   // 4) OFF for 1580us     61
   // 5) ON  for 410us      16
   // 6) OFF for 3580us     138
   // 7) ON  for 400us      15
   // 8) repeat once after 63200us

	// 1) 77 periods ON
   for (i=0; i<77; i++) { 
		one_period_ON();
   }
	// 2) 1070 periods OFF
	for (i=0; i<111; i++) {
	   // x5 OFF
		one_period_OFF();
		one_period_OFF();
		one_period_OFF();
		one_period_OFF();
		one_period_OFF();
		// x5 OFFl
		one_period_OFF();
		one_period_OFF();
		one_period_OFF();
		one_period_OFF();
		one_period_OFF();
	}
	// 3) 15 periods ON
	for (i=0; i<15; i++) {
		one_period_ON();
   }
	// 4) 61 periods OFF
	for (i=0; i<61; i++) {
		one_period_OFF();
   }
	// 5) 16 periods ON
	for (i=0; i<16; i++) {
		one_period_ON();
   }
	// 6) 138 periods OFF
	for (i=0; i<138; i++) {
		one_period_OFF();
   }
	// 7) 15 periods ON
	for (i=0; i<15; i++) {
		one_period_ON();
   }
   return;
}

////////////////////////////////////////////////////////////
// SLEEP BETWEEN SHUTTER RELEASE HALF-SEQUENCE
////////////////////////////////////////////////////////////

void sleep_63ms()
{
	unsigned char i;
   // 8) repeat once after 63200us
	for (i=0; i<84+10; i++) { // 84*150us*5=63.0ms
      _delay_us(150);
      _delay_us(150);
      _delay_us(150);
      _delay_us(150);
      _delay_us(150);
   }
   _delay_us(100);
   _delay_us(100);
}

////////////////////////////////////////////////////////////
// ENTIRE SHUTTER RELEASE CODE SEQUENCE
////////////////////////////////////////////////////////////

void shutter_release()
{

   /* Debug: show Send mode on dbg pins */
	#if DEBUG
   PORTB &= ~(_BV(DBG_LEARN_PIN));
   PORTB |= _BV(DBG_SEND_PIN);
   #endif

	/* No interrupts */
	cli();

	#if 1
   /* Send sequence parts 1-7 */
	shutter_release_1seq();
	#else
	{
		unsigned char i = 0;
		/* Send a test pulse */
	   for (i=0; i<77; i++) { 
			one_period_ON();
	   }
	}
	#endif

   /* Repeat once after 63200us */
	sleep_63ms();

	#if 1
   /* Send sequence parts 1-7 */
	shutter_release_1seq();
	#else
	{
		unsigned char i = 0;
		/* Send a test pulse */
	   for (i=0; i<77; i++) { 
			one_period_ON();
	   }
	}
	#endif

	/* Interrupts back on */
   sei();
}


////////////////////////////////////////////////////////////
// LEARN THE TIMINGS AFTER A RESET
////////////////////////////////////////////////////////////

void learn_timings()
{
   on_time = 1;
	pause_time = 1;

   /* Debug: show learn mode on dbg pins */
	#if DEBUG
   PORTB |= _BV(DBG_LEARN_PIN);
   PORTB &= ~(_BV(DBG_SEND_PIN));
	#endif

   /* Wait for 1st depress (hits GND) */
   while ((PINB & _BV(SHUTTER_PIN)) != 0) {
		sleep_200ms();
	}
	shutter_release();

	/* Wait for 1st release (hits Vcc), count the exposure time */
   while ((PINB & _BV(SHUTTER_PIN)) == 0) {
	   if (on_time<65534) on_time++;
		sleep_200ms();
	}

	/* Bulb: close shutter again */
	if (bulb_mode==1) {
	   shutter_release();
   }

	/* Wait for 2nd depress (hits GND) to get delay between exposures */
   while ((PINB & _BV(SHUTTER_PIN)) != 0) {
	   if (pause_time<65534) pause_time++;
		sleep_200ms();
	}

   /* Debug: show Send mode on dbg pins */
	#if DEBUG
   PORTB &= ~(_BV(DBG_LEARN_PIN));
   PORTB |= _BV(DBG_SEND_PIN);
	#endif

	/* All further releases/depresses are ignored */
	return;
}

////////////////////////////////////////////////////////////
// INTERVAL TIMER MODE
////////////////////////////////////////////////////////////

void run_interval_timer()
{
   unsigned int cnt;

   /* Open shutter, wait */
	shutter_release();
	for (cnt=0; cnt<on_time; cnt++) {
		sleep_200ms();
	}

	/* Bulb: close shutter */
	if (bulb_mode==1)
	   shutter_release();

   /* Wait until next exposure */
	for (cnt=0; cnt<pause_time; cnt++) {
		sleep_200ms();
	}
}

////////////////////////////////////////////////////////////
// INTERVAL TIMER MODE
////////////////////////////////////////////////////////////

int main()
{
   init();

	/* Debug mode */
	#if 0
	while (1) {
		shutter_release();
		sleep_63ms();
	}
	#endif

   learn_timings();
	while(1) {
	   run_interval_timer();
	}
   return 0;
}
